package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.SfaOnlyExperimentModule;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.lowlevel.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.testing.base.*;
import be.iminds.ilabt.jfed.testing.shared.CommonAMTest;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStep;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.testing.shared.ProxyInfoGenerator;
import be.iminds.ilabt.jfed.util.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.Injector;
import io.dropwizard.jackson.Jackson;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.inject.Inject;
import javax.validation.constraints.NotNull;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.Files;
import java.security.NoSuchAlgorithmException;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * GuiLogicTest
 */
public class GuiLogicTest extends ApiTest<GuiLogicTestConfig> {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(GuiLogicTest.class);

    @Nonnull Server userAuthorityServer;

    private final Injector injector;

    @Inject
    public GuiLogicTest(@Nonnull Injector injector,
                        @Nonnull Logger logger,
                        @Nonnull TargetAuthority testedAuthority,
                        @Nonnull GeniUserProvider geniUserProvider,
                        @Nonnull JFedConnectionProvider connectionProvider,
                        @Nonnull TestbedInfoSource testbedInfoSource,
                        @Nonnull AuthorityFinder authorityFinder,
                        @Nonnull JFedPreferences jFedPreferences) {
        super(logger,
                testedAuthority, geniUserProvider,
                connectionProvider, testbedInfoSource,
                authorityFinder, jFedPreferences);

        this.injector = injector;

        if (user == null)
            throw new IllegalStateException("User is null");

        Server uas = user.getUserAuthorityServer();
        if (uas == null)
            throw new IllegalStateException("User has no Authority Server");

        userAuthorityServer = uas;
    }

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final ApiTestMetaData metadata = new ApiTestMetaData<GuiLogicTestConfig>() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "A test using the jFed GUI Editor logic. This will reserve resources, test login, and delete them again.";
        }

        @Nonnull
        @Override
        public Class<GuiLogicTestConfig> getConfigClass() {
            return GuiLogicTestConfig.class;
        }

        @Nonnull
        @Override
        public GuiLogicTestConfig parseApiTestConfig(@Nonnull String apiTestConfigString) {
            try {
                return MAPPER.readValue(apiTestConfigString, GuiLogicTestConfig.class);
            } catch (IOException e) {
                throw new RuntimeException("Invalid Test config", e);
            }
        }

        @Nonnull
        @Override
        public ApiTestConfigEditor<GuiLogicTestConfig> getApiTestConfigEditor() {
            return new JsonTestConfigEditor(GuiLogicTestConfig.class);
        }

    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }

    NodeLoginTestStep nodeLoginTestStep;

    @Override
    public void setUp() throws Exception {
        assertNotNull(geniUserProvider, "No user specified for test");
        assertNotNull(user, "No user specified for test");
        assertNotNull(getTestedAuthority(), "No target server specified for test");
        assertNotNull(getTestConfig(), "no test config");
        assertTrue(getTestConfig().isValid(), "Test config is not valid");
        assertNotEmpty(getTestConfig().getResources(), "Configuration error: no Resource config");
        resourceInfos = getTestConfig().getResources().stream().filter(Objects::nonNull).map(r -> new ResourceInfo(r)).collect(Collectors.toList());
        assertEquals(resourceInfos.size(), getTestConfig().getResources().size(), "Configuration error: some of the Resources in the config where null");

        assertNotEmpty(resourceInfos, "bug");

        this.nodeLoginTestStep =
                new NodeLoginTestStep(this, getTestConfig().generateNodeLoginTestStepConfig(geniUserProvider));

        //fail early
        for (ResourceInfo resourceInfo : resourceInfos) {
            assertTrue(resourceInfo.config.getProvision().isEnabled(),
                    "A disabled provision step is currently not supported.");
        }

        setupAnsible();
    }

    /** Info on a Resource that is shared between test methods. */
    private class ResourceInfo {
        @NotNull private final GuiLogicTestConfig.Resource config;

        private GeniUrn sliceUrn;

        private String requestRspec;
        private int maxNodeNameLength = 5; //TODO: automatically get from rspec
        private List<Server> serversInRspec;
        private RequestRspecSource requestRspecSource;
        public String sliceName;

        public ResourceInfo(@NotNull GuiLogicTestConfig.Resource config) {
            this.config = config;
        }

        private String project;
        private Experiment experiment;
        private boolean notEnoughResourcesDetected = false;
        private ExperimentController experimentController;

        private boolean waitForExperimentIsOver = false;
        private boolean deleted = false;

        String manifestRspecString;
        boolean validManifest;

        public boolean isWaitForExperimentOver() {
            return waitForExperimentIsOver;
        }

        public boolean isDeleted() {
            return deleted;
        }
    }

    private List<ResourceInfo> resourceInfos;

    @Test(groups = {"init"})
    public void getRspec() throws GeniUrn.GeniUrnParseException {
        int resourceIndex = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            GuiLogicTestConfig.RequestRSpec rspecConfig = resourceInfo.config.getRequestRSpec();
            switch (rspecConfig.getSource()) {
                case GENERATE_SIMPLE: {
                    @Nonnull String exclusive = rspecConfig.getNodeExclusive() ? "true" : "false";
                    @Nonnull String sliverTypeName = rspecConfig.getNodeSliverType();
                    int nodeCount = rspecConfig.getNodeCount();
                    @Nonnull List<String> compIds = rspecConfig.getNodeComponentUrn();
                    assertTrue(compIds.isEmpty() || nodeCount == compIds.size(), "Config error: Mismatch of componentIds and nodeCount");
                    @Nonnull String compManId = getTestedAuthority().getServerForRspecComponentManager().getDefaultComponentManagerUrn();
                    resourceInfo.requestRspec = "<?xml version='1.0'?>\n" +
                            "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"request\"  " +
                            "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                            "xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n";

                    for (int i = 0; i < nodeCount; i++) {
                        @Nonnull String compId = "";
                        if (compIds.size() > i)
                            compId = " component_id=\""+compIds.get(i)+"\"";
                        resourceInfo.requestRspec +=
                                "  <node client_id=\"n"+i+"\" exclusive=\"" + exclusive + "\" component_manager_id=\"" + compManId + "\""+compId+">\n" +
                                        "    <sliver_type name=\"" + sliverTypeName + "\"/>\n" +
                                        "  </node>\n";
                    }

                    resourceInfo.requestRspec += "</rspec>";

                    note("Generated Request RSpec using \"simple\" method:");
                    note(resourceInfo.requestRspec, true);
                    break;
                }
                case PROVIDE_URL: {
                    String rspecUrl = rspecConfig.getProvidedContentSource();
                    assertNotNull(rspecUrl, "Config error: no RSpec URL provided");

                    URL url;
                    try {
                        url = new URL(rspecUrl);
                    } catch (MalformedURLException e) {
                        throw new RuntimeException("Invalid URL specified for RSpec: \""+rspecUrl+"\"", e);
                    }
                    try (InputStream is = url.openStream()) {
                        String rspec = IOUtils.streamToString(is, "UTF-8");
                        if (rspec.trim().isEmpty()) {
                            warn("Empty Rspec read from URL \"" + rspecUrl + "\"");
                        }

                        String head = TextUtil.abbreviate(rspec, 200);
                        note("Downloaded Rspec from \"" + rspecUrl + "\". Rspec size: " + rspec.length() + " characters.\nHead: " + head, true/*formatAsCode*/);
                        resourceInfo.requestRspec = rspec;
                    } catch (IOException e) {
                        throw new RuntimeException("Failed to fetch RSpec at \""+rspecUrl+"\"", e);
                    }
                    break;
                }
                case PROVIDE_CONTENT: {
                    resourceInfo.requestRspec = rspecConfig.getProvidedContentSource();
                    assertNotNull(resourceInfo.requestRspec, "Config error: no RSpec provided");

                    String head = TextUtil.abbreviate(resourceInfo.requestRspec, 200);
                    note("Using Request Rspec from config. " +
                            "Size: " + resourceInfo.requestRspec.length() + " characters.\n" +
                            "Head: " + head, true/*formatAsCode*/);

                    break;
                }
                default: throw new RuntimeException("RequestRSpec source "+resourceInfo.config.getRequestRSpec().getSource()+" not yet implemented");
            }

            if (rspecConfig.getBindTestedServerToUnboundNodes()) {
                String origRequestRspec = resourceInfo.requestRspec;
                String targetServerCM = testedAuthority.getServerForRspecComponentManager().getDefaultComponentManagerUrn();
                resourceInfo.requestRspec = CommonAMTest.bindUnboundTo(this, resourceInfo.requestRspec, targetServerCM);
                if (!origRequestRspec.equals(resourceInfo.requestRspec)) {
                    note("Unbound nodes were bound to "+targetServerCM);
                }
            }

            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("rspec"+resourceIndex)
                            .setValue(resourceInfo.requestRspec)
                            .setLarge(true)
                            .setHttpMediaTypeToXml()
            );

            resourceInfo.requestRspecSource = new RequestRspecSource(resourceInfo.requestRspec, ModelRspecType.BASIC);
            resourceInfo.serversInRspec = resourceInfo.requestRspecSource.getAllComponentManagerUrns().stream()
                    .map(serverUrn ->
                            authorityFinder.findByUrn(
                                    serverUrn,
                                    AuthorityFinder.Purpose.REQUEST_RSPEC)
                    )
                    .collect(Collectors.toList());

            resourceInfo.maxNodeNameLength = resourceInfo.requestRspecSource.getBasicNodeInfo().stream()
                    .map(BasicStringRspec.BasicNodeInfo::getClientId)
                    .filter(Objects::nonNull)
                    .mapToInt(String::length)
                    .max()
                    .orElse(5); //uses 5 as fallback when no nodes with client_id found

            resourceIndex++;
        }
    }

    private UserAndSliceApiWrapper userAndSliceApiWrapper;
    @Test(hardDepends = {"getRspec"}, groups = {"init"})
    public void findProject() {
        assertNotNull(user);
        userAndSliceApiWrapper =
                this.getAutomaticUserAndSliceApiWrapperFactory().create();

        for (ResourceInfo resourceInfo : resourceInfos) {
            switch (resourceInfo.config.getSlice().getProjectSource()) {
                case NONE: {
                    resourceInfo.project = null;
                    break;
                }
                case PROVIDED: {
                    resourceInfo.project = resourceInfo.config.getSlice().getProject();
                    break;
                }
                case AUTO: {
                    resourceInfo.project = CommonAMTest.findProject(
                            this,
                            this.getLogger(),
                            resourceInfo.project,
                            user,
                            user.getUserUrn(),
                            userAndSliceApiWrapper,
                            true);
                    break;
                }
            }

            if (resourceInfo.project != null) {
                note("Will use project: \"" + resourceInfo.project + "\"");
            } else {
                assertFalse(resourceInfo.config.getSlice().getFailIfNoProject(), "Empty project is not allowed by config");
                note("Will NOT use a project");
            }
        }
    }

    @Test(hardDepends = {"getRspec", "findProject"}, groups = {"init"})
    public void createExperiment() {
        assertNotNull(user);
        for (ResourceInfo resourceInfo : resourceInfos) {
            resourceInfo.sliceName = CommonAMTest.createSliceName(
                    this,
                    userAuthorityServer,
                    ApiInfo.getDefaultComponentManagerForEach(resourceInfo.serversInRspec),
                    "g",
                    resourceInfo.maxNodeNameLength,
                    null);

            note("Will create slice \"" + resourceInfo.sliceName + "\"");

            Instant requestedEndTime = Instant.now().plus(resourceInfo.config.getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
            note("Created slice (and slivers) will expire at " + requestedEndTime + "");

//        List<String> sshKeys = Collections.singletonList(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey()));
            List<String> sshKeys = Collections.singletonList(nodeLoginTestStep.getSshKeyHelper().getSshPublicKeyString());
            UserSpec userSpec = new UserSpec(user.getUserUrnString(), sshKeys);
            List<UserSpec> userSpecs = Collections.singletonList(userSpec);

            resourceInfo.experiment = new Experiment(
                    resourceInfo.sliceName,
                    resourceInfo.project,
                    resourceInfo.requestRspecSource,
                    null,
                    requestedEndTime,
                    userSpecs,
                    false);
        }
    }

    private Injector highLevelInjector;
    @Test(hardDepends = {"getRspec"}, groups = {"run"})
    public void createInjector() {
        List<com.google.inject.AbstractModule> extraMods = new ArrayList<>();

        boolean hasHighLevelModule = false;
        boolean hasSfaOnlyExperimentModule = false;

        try {
            hasHighLevelModule = injector.getInstance(ApiCallDetailsCache.class) != null;
        } catch (Exception e) {
            LOG.info("Binding from HighLevelModule does not exists "+e.getClass().getName(), e);
        }
        try {
            hasSfaOnlyExperimentModule = injector.getInstance(ExperimentPartControllerManager.class) != null;
        } catch (Exception e) {
            LOG.info("Binding from SfaOnlyExperimentModule does not exists "+e.getClass().getName(), e);
        }

        if (!hasHighLevelModule) {
            extraMods.add(new HighLevelModule());
            note("Will add HighLevelModule to injector");
        } else {
            note("Will NOT add HighLevelModule to injector");
        }
        if (!hasSfaOnlyExperimentModule) {
            extraMods.add(new SfaOnlyExperimentModule());
            note("Will add SfaOnlyExperimentModule to injector");
        } else {
            note("Will NOT add SfaOnlyExperimentModule to injector");
        }

//        AbstractModule loggerOverride = new AbstractModule() {
//            @Override
//            protected void configure() {
//                todo;
//            }
//        };
//
//        todo; Modules.override(extraMods).with(loggerOverride);

        highLevelInjector = injector.createChildInjector();
        assertNotNull(highLevelInjector);
        note("Created injector");

        Logger injectorLogger = injector.getInstance(Logger.class);
        //noinspection ObjectEquality
        if (getLogger() != injectorLogger) {
            errorFatal("Invalid logger in injector");
        } else {
            note("Logger OK");
        }
    }

    @Test(hardDepends = {"createExperiment", "createInjector"}, groups = {"run"})
    public void runExperiment() {
        for (ResourceInfo resourceInfo : resourceInfos) {
            assertTrue(resourceInfo.config.getProvision().isEnabled(),
                    "A disabled provision step is currently not supported.");

            resourceInfo.notEnoughResourcesDetected = false;

            final ExperimentControllerFactory experimentControllerFactory =
                    highLevelInjector.getInstance(ExperimentControllerFactory.class);
            assertNotNull(experimentControllerFactory);

            resourceInfo.experimentController =
                    experimentControllerFactory.createExperimentController(resourceInfo.experiment);
            assertNotNull(resourceInfo.experimentController);

//        experimentController.setExperimentTester(experimenterJobFactory::createTestConnectivityJob);
//        experimentController.addListener(bugReportExperimentControllerListener);

            resourceInfo.experimentController.start();
            note("Started experimentController. ExperimentState is now " + resourceInfo.experiment.getExperimentState());

        }

        //TODO: if not enough free resources detected, do not wait
//        resourceInfo.notEnoughResourcesDetected = todo;

        //Now wait.
        //This cannot be in a seperate method: we might miss calls in between the 2 methods otherwise (because experimentController is async).

        long maxTimeToWaitForReadyInMinutes = 0;
        long extraWaitTimeInSeconds = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (!resourceInfo.config.getWaitForReady().isEnabled()) {
                //don't take this wait and extra time into account

                //consider the wait over
                resourceInfo.waitForExperimentIsOver = true;
                continue;
            }
            if (resourceInfo.config.getWaitForReady().getMaxTimeMin() > maxTimeToWaitForReadyInMinutes) {
                maxTimeToWaitForReadyInMinutes = resourceInfo.config.getWaitForReady().getMaxTimeMin();
            }
            if (resourceInfo.config.getWaitForReady().getExtraWaitTimeSeconds() > extraWaitTimeInSeconds) {
                extraWaitTimeInSeconds = resourceInfo.config.getWaitForReady().getExtraWaitTimeSeconds();
            }
        }

        if (maxTimeToWaitForReadyInMinutes > 0) {
            note("Starting to wait for experiment(s) to become ready at " + new Date());
        } else {
            note("Will not wait for experiment(s) to become ready.");
        }

        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForReadyInMinutes * 60 * 1000L);
        boolean allDone = false;
        boolean sawFailure = false;
        while (now < deadline && !allDone && !sawFailure) {
            for (ResourceInfo resourceInfo : resourceInfos) {
                if (!resourceInfo.waitForExperimentIsOver) {
                    assertTrue(resourceInfo.config.getWaitForReady().isEnabled(), "bug in runExperiment step code.");
                    note("Current ExperimentState=" + resourceInfo.experiment.getExperimentState());
                    note("Checking if experiment is ready.");

                    if (resourceInfo.experiment.getSliceUrn() != null) {
                        resourceInfo.sliceUrn = resourceInfo.experiment.getSliceUrn();
                    }

                    if (resourceInfo.experiment.getExperimentState() == ExperimentState.READY) {
                        note("Ready, because ExperimentState is READY");
                        resourceInfo.waitForExperimentIsOver = true;
                    }
                    if (resourceInfo.experiment.getExperimentState().isTerminated()) {
                        resourceInfo.waitForExperimentIsOver = true;
                        sawFailure = true;
                        errorNonFatal("Failure, because ExperimentState is terminated");
                    }
                    if (!resourceInfo.experiment.getExperimentState().isActiveState()) {
                        //TODO detect and set notEnoughResourcesDetected
                        resourceInfo.waitForExperimentIsOver = true;
                        sawFailure = true;
                        errorNonFatal("Failure, because ExperimentState is not active state");
                    }
                }
            }

            allDone = resourceInfos.stream().map(ResourceInfo::isWaitForExperimentOver).allMatch(b -> b == true);

            if (!allDone && !sawFailure) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }

        if (maxTimeToWaitForReadyInMinutes > 0) {
            note("Stopped waiting for experiment(s) to become ready at " + new Date());
        }

        if (extraWaitTimeInSeconds > 0) {
            note("Waiting "+extraWaitTimeInSeconds+" extra seconds before continuing. (due to resource.waitForReady.extraWaitTimeSeconds option)");
            try {
                Thread.sleep(extraWaitTimeInSeconds * 1000L);
            } catch (InterruptedException e) {
                warn("Extra waiting time was interrupted. This should not happen.", e);
            }
        }

        if (sawFailure) {
            errorFatal("One or more experiments have failed.");
        }

        if (!allDone) {
            errorFatal("One or more experiments did not become ready by the deadline.");
        }
    }

    public boolean isValidGeniRspec(String rspec) {
        return new BasicStringRspec(rspec).isValidRspec();
    }


    @Test(hardDepends = {"runExperiment"}, groups = {"login"})
    public void checkManifest() throws JFedException {
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.notEnoughResourcesDetected) {
                warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
                return;
            }

            Slice slice = resourceInfo.experiment.getSlice();
            assertNotNull(slice, "experiment slice should not be null");
            ManifestRspecSource manifestRspecSource = slice.getManifestRspec();
            assertNotNull(manifestRspecSource, "experiment slice manifestRspecSource should not be null");
            resourceInfo.manifestRspecString =  manifestRspecSource.getRspecXmlString();
            assertNotNull(resourceInfo.manifestRspecString, "experiment slice manifestRspecSource getRspecXmlString should not be null");

            boolean validRequest = isValidGeniRspec(resourceInfo.requestRspec);
            resourceInfo.validManifest = isValidGeniRspec(resourceInfo.manifestRspecString);

            if (resourceInfo.requestRspec != null && validRequest && resourceInfo.validManifest) {
                setErrorsNotFatal();
                assertTrue(CommonAMTest.sameNodesInRequestAndManifest(this, resourceInfo.requestRspec, resourceInfo.manifestRspecString));
                setErrorsFatal();
            }

            if (resourceInfo.validManifest) {
                boolean foundNodeLogin = nodeLoginTestStep.parseSshInfoFromGeni3ManifestRspec(resourceInfo.manifestRspecString, null, null);
                if (foundNodeLogin)
                    note("Successfully found node login in manifest");
                else
                    note("Did not find node login info in manifest");

                if (resourceInfo.config.getWaitForReady().getSkipFinalManifestCheck())
                    nodeLoginTestStep.checkManifestCorrectness(resourceInfo.manifestRspecString, testedAuthority);
            }
        }
    }

    @Test(hardDepends = {"checkManifest"}, groups = {"login"})
    public void login() throws NoSuchAlgorithmException, IOException {
        if (!getTestConfig().getNodeLoginTest().isEnabled()) {
            skip("Test skipped because disabled in config");
            return;
        }

        //todo; //TODO use nodes from different RSpecs for node login test
        //todo; //TODO log in to multiple nodes

        assertNotNull(geniUserProvider, "No user specified for test");
        assert nodeLoginTestStep != null;

        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.notEnoughResourcesDetected) {
                warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
                return;
            }
            if (!resourceInfo.validManifest) {
                skip("Test skipped because at least one manifest is not a geni version 3 RSpec (so this test does not know how to extract node login info)");
                return;
            }
        }

        String sliceName = resourceInfos.get(0).sliceName;

        nodeLoginTestStep.testNodeLogin(false/*only note, no warning*/, geniUserProvider, sliceName);
    }

    private JFedConnection.SshProxyInfo convertProxy(be.iminds.ilabt.jfed.testing.shared.Proxy proxy, ResourceInfo resourceInfo) {
        assertNotNull(geniUserProvider, "No user specified for test");
        return ProxyInfoGenerator.generate(this, proxy, geniUserProvider, resourceInfo.sliceName, nodeLoginTestStep.getSshKeyHelper(), nodeLoginTestStep.getSshUsername());
    }


    @Test(hardDepends = {"checkManifest"}, groups = {"login"})
    public void ansible() throws NoSuchAlgorithmException, IOException {
        if (!getTestConfig().getAnsibleTest().isEnabled()) {
            skip("Test skipped because disabled in config");
            return;
        }

        LOG.debug("Starting ansible test");
        JFedConnection.SshProxyInfo sshProxy = convertProxy(getTestConfig().getAnsibleTest().getProxy(), resourceInfos.get(0));

        assertTrue(sshProxy == null, "ansible does not support an SSH proxy, but proxy configured: " + sshProxy);

        try {
            writeAnsibleFiles(getTestConfig().getAnsibleTest().getDebug());
            LOG.debug("   wrote ansible files to " + ansibleDir);
        } catch (IOException e) {
            errorFatal("Failed to write ansible files", e);
        }
        String output = callAnsible();
        LOG.debug("    got ansible output");

        if (output != null) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "output")
                            .setValue(output)
                            .setLarge(true)
                            .setHttpMediaTypeToText());
        }

        if (output == null || output.trim().isEmpty())
            fatalError("Ansible did not return any output.");

        note("Ansible output:\n" + output + "\n", true/*formatAsCode*/);

        checkAnsibleOutput(
                output, this,
                null, null, null,
                getTestConfig().getAnsibleTest().getSuccessRegex(),
                getTestConfig().getAnsibleTest().getWarningRegex(),
                getTestConfig().getAnsibleTest().getFailureRegex());
    }

    @Test(softDepends = {"login", "ansible", "runExperiment"},
            hardDepends = {"getRspec", "createInjector", "createExperiment"},
            groups = {"cleanup", "login", "run"})
    public void deleteExperiment() {
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.config.getCleanup().isEnabled()) {
                note("Will call experimentController.stop();");
                resourceInfo.experimentController.stop();
            } else {
                note("Experiment Cleanup disabled");
                resourceInfo.deleted = true;
            }
        }

        long maxTimeToWaitForDeleteInMinutes = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.config.getCleanup().getMaxWaitTimeMinutes() > maxTimeToWaitForDeleteInMinutes) {
                maxTimeToWaitForDeleteInMinutes = resourceInfo.config.getCleanup().getMaxWaitTimeMinutes();
            }
        }

        note("Will wait until experiment is terminated");
        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForDeleteInMinutes * 60 * 1000L);
        boolean allDeleted = false;
        while (now < deadline && !allDeleted) {
            for (ResourceInfo resourceInfo : resourceInfos) {
                if (!resourceInfo.deleted) {
                    note("Current ExperimentState=" + resourceInfo.experiment.getExperimentState());
                    if (resourceInfo.experiment.getExperimentState() == ExperimentState.EMPTY ||
                            resourceInfo.experiment.getExperimentState() == ExperimentState.EXPIRED ||
                            resourceInfo.experiment.getExperimentState() == ExperimentState.FAILED) {
                        note("Experiment was terminated");
                        resourceInfo.deleted = true;
                    }
                }
            }

            allDeleted = resourceInfos.stream().map(ResourceInfo::isDeleted).allMatch(b -> b == true);

            if (!allDeleted) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }

        if (!allDeleted) {
            errorFatal("Experiment(s) did not terminate within " + maxTimeToWaitForDeleteInMinutes + " minutes. Gave up waiting for it.");
        } else {
            note("All experiments deleted");
        }
    }




    private File ansibleDir;
    private File ansiblePlaybookLocalFile;

    private String ansiblePlaybookExeName;
    private String ansiblePlaybookContent;

    public void setupAnsible() {
        GuiLogicTestConfig.AnsibleTest config = getTestConfig().getAnsibleTest();
        if (config.isEnabled()) {
            //isValid of the config checks a lot already, so we can assume a lot here

            ansiblePlaybookExeName = getTestConfig().getAnsibleTest().getPlaybookExe().trim();

            if (config.getPlaybookContent() != null) {
                ansiblePlaybookContent = config.getPlaybookContent();
            } else {
                URL url;
                try {
                    //noinspection ConstantConditions
                    url = new URL(config.getPlaybookUrl());
                } catch (MalformedURLException e) {
                    throw new RuntimeException("Invalid URL specified for RSpec: \""+config.getPlaybookUrl()+"\"", e);
                }
                try (InputStream is = url.openStream()) {
                    ansiblePlaybookContent = IOUtils.streamToString(is, "UTF-8");
                    if (ansiblePlaybookContent.trim().isEmpty()) {
                        warn("Empty Playbook read from URL \"" + ansiblePlaybookContent + "\"");
                    }

                    String head = TextUtil.abbreviate(ansiblePlaybookContent, 200);
                    note("Downloaded Playbook from \"" + config.getPlaybookUrl() + "\". Playbook size: " + ansiblePlaybookContent.length() + " characters.\n" +
                                    "Head: " + head,
                            true/*formatAsCode*/);
                } catch (IOException e) {
                    throw new RuntimeException("Failed to fetch Playbook at \""+config.getPlaybookUrl()+"\"", e);
                }
            }

            try {
                ansibleDir = Files.createTempDirectory("ansibleTest").toFile();
            } catch (IOException e) {
                errorFatal("Could not create temporary dir for ansible files.", e);
            }

            ansiblePlaybookLocalFile = new File(ansibleDir, "playbook.yml");

            try {
                FileWriter fw = new FileWriter(ansiblePlaybookLocalFile);
                fw.write(ansiblePlaybookContent);
                fw.close();
            } catch (IOException e) {
                errorFatal("Error writing '" + ansiblePlaybookLocalFile.getPath() + "': " + e.getMessage(), e);
            }

            if (!(ansibleDir.exists() && ansibleDir.isDirectory())) {
                File parentDir = ansibleDir.getParentFile();
                if ((parentDir != null && !ansibleDir.exists()) && parentDir.exists() && parentDir.isDirectory()) {
                    //noinspection ResultOfMethodCallIgnored
                    ansibleDir.mkdir();
                    if (!(ansibleDir.exists() && ansibleDir.isDirectory()))
                        errorFatal("The ansible directory \"" + ansibleDir.getPath()+ "\" does not exist and could be created.");
                } else
                    errorFatal("The ansible directory \"" + ansibleDir.getPath() + "\" does not exist and will not be created because the parent dir doesn't exist either.");
            }
            if (!(ansiblePlaybookLocalFile.exists() && ansiblePlaybookLocalFile.isFile() && ansiblePlaybookLocalFile.canRead())) {
                errorFatal("The ansible playbook file \"" + ansiblePlaybookLocalFile.getPath() + "\" does not exist (or is not readable)");
            }

            note("Preparation for ansible test succesfull");
        }
    }

    static void checkAnsibleOutput(String output,
                                   ApiTest test,
                                   String ansibleSuccessWord, String ansibleWarningWord, String ansibleFailureWord,
                                   String ansibleSuccessRegex, String ansibleWarningRegex, String ansibleFailureRegex) {
        if (ansibleFailureWord != null && output.contains(ansibleFailureWord))
            test.fatalError("Error: Ansible output contained \"" + ansibleFailureWord + "\"");
        if (ansibleFailureRegex != null && Pattern.compile(ansibleFailureRegex).matcher(output).find())
            test.fatalError("Error: Ansible output matched regex \"" + ansibleFailureRegex + "\"");
        if (ansibleWarningWord != null && output.contains(ansibleWarningWord))
            test.warn("Warning: Ansible output contained \"" + ansibleWarningWord + "\"");
        else {
            if (ansibleWarningRegex != null && Pattern.compile(ansibleWarningRegex).matcher(output).find())
                test.warn("Warning: Ansible output matched regex \"" + ansibleWarningRegex + "\"");
            else {
                if (ansibleSuccessWord != null && output.contains(ansibleSuccessWord))
                    test.note("Success: Ansible output contained \"" + ansibleSuccessWord + "\"");
                else {
                    if (ansibleSuccessRegex != null && Pattern.compile(ansibleSuccessRegex).matcher(output).find())
                        test.note("Success: Ansible output matched regex \"" + ansibleSuccessRegex + "\"");
                    else {
                        if (ansibleSuccessWord != null) {
                            if (ansibleSuccessRegex != null)
                                test.fatalError("Error: Ansible output did not contain \"" + ansibleSuccessWord + "\" or match regex \"" + ansibleSuccessRegex + "\"");
                            else
                                test.fatalError("Error: Ansible output did not contain \"" + ansibleSuccessWord + "\"");
                        } else
                            test.fatalError("Error: Ansible output did not match regex \"" + ansibleSuccessRegex + "\"");
                    }
                }
            }
        }
    }

    private void writeAnsibleFiles(boolean addDebug) throws IOException {
        RspecFactory rspecFactory = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
        ResourceInfo resourceInfo = resourceInfos.get(0);
        assertNotNull(resourceInfo.requestRspecSource, "No Request Rspec for resource 0");
        ModelRspec first = resourceInfo.requestRspecSource.getModelRspec(ModelRspecType.BASIC);
        if (first == null) {
            fatalError("Failed to parse request rspec 0");
        }
        ModelRspec mergedRequestRspec = rspecFactory.copyModelRspec(first, "request");

        for (int i = 1; i < resourceInfos.size(); i++) {
            resourceInfo = resourceInfos.get(i);
            assertNotNull(resourceInfo.requestRspecSource, "No Request Rspec for resource " + i);

            ModelRspec cur = resourceInfo.requestRspecSource.getModelRspec(ModelRspecType.BASIC);
            if (cur == null) {
                fatalError("Failed to parse request rspec " + i);
            }

            //copy all nodes to the merged RSpec. (Don't care about links and interfaces.)
            for (RspecNode node : cur.getNodes()) {
                RspecNode mergedNode = rspecFactory.copyNode(mergedRequestRspec, node, RspecNode.InterfaceCopyMethod.NO_COPY);
                mergedRequestRspec.addNode(mergedNode);
                if (addDebug) {
                    note("Added node to merged request RSpec: "+node.getClientId());
                }
            }
        }

        String mergedRequestRspecString = mergedRequestRspec.toGeni3Rspec();
        if (addDebug) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "mergedRequestRspec")
                            .setValue(mergedRequestRspecString)
                            .setLarge(true)
                            .setHttpMediaTypeToXml()
            );
        }

        //this uses the alternate manifests of AnsibleFileWriter to get all possible logins from all received manifests

        //safe to copy private key, as it is generated just for this test
        AnsibleFileWriter ansibleFileWriter =
                AnsibleFileWriter.createWithCopiedPrivateKey(
                        new BasicStringRspec(mergedRequestRspecString),
                        nodeLoginTestStep.getSshKeyHelper().getSshPrivateKey(),
                        nodeLoginTestStep.getSshKeyHelper().getSshPublicKey(),
                        getUser(),
                        null,
                        nodeLoginTestStep.getSshUsername()
                );


        for (ResourceInfo curResourceInfo : resourceInfos) {
            assert curResourceInfo.manifestRspecString != null;
            if (curResourceInfo.manifestRspecString != null) {
                ansibleFileWriter.addAltBasicStringRspec(curResourceInfo.manifestRspecString);
            }
        }

        ansibleFileWriter.writeFilesToDir(ansibleDir);
        if (addDebug) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "hostsFile")
                            .setValue(ansibleFileWriter.getAnsibleHostFileContent())
                            .setLarge(true)
                            .setHttpMediaTypeToText()
            );
        }
    }

    private String callAnsible() {
        ArrayList<String> command = new ArrayList<>();
        command.add(ansiblePlaybookExeName);
        command.add(ansiblePlaybookLocalFile.getAbsolutePath());
        if (getTestConfig().getAnsibleTest().getDebug())
            command.add("-vvvv");
        try {
            ProcessBuilder pb = new ProcessBuilder(command);
            pb.directory(ansibleDir);
            pb.redirectErrorStream(); //combine stderr and stout, so we can read both at once in one thread. Makes everything easy.

            Process p = pb.start();
            BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

            String output = "";
            String line;
            while ((line = input.readLine()) != null) {
                output += line + System.lineSeparator();
            }
            //note: input might close before terminal stops.
            input.close();

            try {
                p.waitFor();
            } catch (InterruptedException e) {
                //ignore
            }

            return output;
        } catch (Exception e) {
            fatalError("Error trying to call ansible with command: " + command + "", e);
            return null;
        }
    }
}
